from .loader import BufferLoader
from .loader import Loader
from .loader import FROM_STDIN
from .loader import as_loader
from .loader import load
from .loader import load_json
from .loader import wrapopen
